/* view.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2001 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include "giram.h"
#include "tool_select.h"
#include "tool_move.h"
#include "tool_rotate.h"
#include "tool_scale.h"
#include "tool_options.h"
#include "box.h"
#include "cone.h"
#include "csg.h"
#include "cylinder.h"
#include "disc.h"
#include "mesh.h"
#include "plane.h"
#include "sor.h"
#include "sphere.h"
#include "superellipsoid.h"
#include "torus.h"
#include "triangle.h"
#include "light_source.h"
#include "object.h"
#include "csgtree.h"
#include "viewXY.h"
#include "viewXZ.h"
#include "viewZY.h"
#include "utils.h"
#include "camera.h"
#include "SavePov.h"
#include "plugins.h"
#include "edit.h"
#include "snap.h"
#include "dxf.h"
#include "gtkcolorbutton.h"
#include "texture.h"
#ifdef WITH_S3D
#include "s3d.h"
#endif
#include "gts.h"

#include "giramwidgets.h"
#include "giramrc.h"
#include "giramintl.h"

#include "pixmaps/navbutton.xpm"

static gchar *g_stpcpy(gchar *dest, const gchar *src)
{
  register       gchar *d = dest;
  register const gchar *s = src;

  g_return_val_if_fail (dest != NULL, NULL);
  g_return_val_if_fail (src != NULL, NULL);
  do
    *d++ = *s;
  while (*s++ != '\0');

  return d - 1;
}

void giram_view_set_title(ViewStruct *view)
{
  gchar     *title, *ptr;
  GtkWidget *canvas, *window;
  gchar     *short_name, *long_name, *style = "";
  gint       length, i;

  canvas = view->drawing_area;
  window = gtk_widget_get_toplevel(canvas);

  if (view->frame->file_name)
  {
    long_name = view->frame->file_name;
    short_name = g_basename(long_name);
  } else
  {
    long_name = short_name = _("Untitled");
  }

  switch (view->camera_style)
  {
    case ORTHO_XY_CAMERA:
      style = "X-Y";
      break;

    case ORTHO_XZ_CAMERA:
      style = "X-Z";
      break;

    case ORTHO_ZY_CAMERA:
      style = "Z-Y";
      break;
  }

  if (view_title_format == NULL)
    gtk_window_set_title(GTK_WINDOW(window), "Giram");

  length = 0;
  for (i = 0 ; view_title_format[i] ; i++)
  {
    if (view_title_format[i] == '%')
    {
      switch (view_title_format[i+1])
      {
        case 't': /* camera style */
          i++; length += strlen(style);
          break;
        case 's': /* short file name */
          i++; length += strlen(short_name);
          break;
        case 'f': /* full file name */
          i++; length += strlen(long_name);
          break;
        case '%': /* the percent char */
          i++; length++;
          break;
        case 0:
          g_message(_("The view title format string ended within %%-sequence"));
          break;
        default:
          i++;
          g_message(_("The view title format string contains unknown format sequence '%%%c'"),
                      view_title_format[i]);
      }
    } else
      length++;
  }

  ptr = title = g_malloc(length+1);
  for (i = 0 ; view_title_format[i] ; i++)
  {
    if (view_title_format[i] == '%')
    {
      switch (view_title_format[i+1])
      {
        case 't': /* camera style */
          i++; ptr = g_stpcpy(ptr, style);
          break;
        case 's': /* short file name */
          i++; ptr = g_stpcpy(ptr, short_name);
          break;
        case 'f': /* long file name */
          i++; ptr = g_stpcpy(ptr, long_name);
          break;
        case '%':
          i++; *ptr++ = '%';
          break;
        case 0:
          break;
        default:
          *ptr++ = '%';
          i++;
          *ptr++ = view_title_format[i];
      }
    } else
      *ptr++ = view_title_format[i];
  }
  *ptr = 0;

  gtk_window_set_title(GTK_WINDOW(window), title);
}

/*****************************************************************************
*  SetWireFrameExpose
******************************************************************************/
void SetWireFrameExpose(void)
{ /* FIXME: OO me */
  ViewStruct *view_data;
  GtkWidget  *canvas;

  view_data = get_current_view_data();
  canvas = view_data->drawing_area;
  gtk_signal_disconnect(GTK_OBJECT(canvas), view_data->render_style);

  switch (view_data->camera_style)
  {
    case ORTHO_XY_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXY, NULL);
      break;

    case ORTHO_XZ_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXZ, NULL);
      break;

    case ORTHO_ZY_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(canvas);
}

/*****************************************************************************
*  SetHiddenFacesExpose
******************************************************************************/
void SetHiddenFacesExpose(void)
{ /* FIXME: OO me */
  ViewStruct *view_data;
  GtkWidget  *canvas;

  view_data = get_current_view_data();
  canvas = view_data->drawing_area;
  gtk_signal_disconnect(GTK_OBJECT(canvas), view_data->render_style);

  switch (view_data->camera_style)
  {
    case ORTHO_XY_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewXY, NULL);
      break;

    case ORTHO_XZ_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewXZ, NULL);
      break;

    case ORTHO_ZY_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeHiddenFacesViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(canvas);
}

/*****************************************************************************
*  SetGouraudExpose
******************************************************************************/
void SetGouraudExpose(void)
{ /* FIXME: OO me */
  ViewStruct *view_data;
  GtkWidget  *canvas;

  view_data = get_current_view_data();
  canvas = view_data->drawing_area;
  gtk_signal_disconnect(GTK_OBJECT(canvas), view_data->render_style);

  switch (view_data->camera_style)
  {
    case ORTHO_XY_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewXY, NULL);
      break;

    case ORTHO_XZ_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewXZ, NULL);
      break;

    case ORTHO_ZY_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeGouraudViewZY, NULL);
      break;
  }
  gtk_widget_queue_draw(canvas);
}

/*****************************************************************************
*  CallPovFront
******************************************************************************/
void CallPovFront(void)
{
  ViewStruct  *view_data;
  FrameStruct *LocalFrame;
  gchar       *TmpFileName;
  gchar       *CommandLine;

  TmpFileName = g_strdup("/tmp/GiramXXXXXX");
  view_data = get_current_view_data();
  LocalFrame = view_data->frame;
  close(mkstemp(TmpFileName));
  SavePovByName(LocalFrame, TmpFileName);
  CommandLine = g_strdup_printf("povfront -p -s %s &", TmpFileName);
  system(CommandLine);
  g_free(CommandLine);
  g_free(TmpFileName);
}

/*****************************************************************************
*  ZoomIn
******************************************************************************/
void ZoomIn(void)
{
  ViewStruct *view_data;

  view_data = get_current_view_data();

  view_data->zoom *= 1.5;
  gtk_widget_queue_draw(GTK_WIDGET(view_data->drawing_area));
  /* FIXME: Rulers and scrollers */
}

/*****************************************************************************
*  ZoomOut
******************************************************************************/
void ZoomOut(void)
{
  ViewStruct *view_data;

  view_data = get_current_view_data();

  view_data->zoom /= 1.5;
  gtk_widget_queue_draw(GTK_WIDGET(view_data->drawing_area));
  /* FIXME: Rulers and scrollers */
}

/*****************************************************************************
*  FitToSelectionFunc
******************************************************************************/
void FitToSelectionFunc(void)
{
  ViewStruct         *view_data;
  GList              *tmp_select;
  ObjectStruct       *object;
  TriangleListStruct *TmpTri;
  Vector              MaxExt, MinExt;
  double              ZoomX=1.0, ZoomY=1.0;
  GtkWidget          *canvas;

  view_data = get_current_view_data();
  canvas = view_data->drawing_area;

  MinExt[0] = MinExt[1] = MinExt[2] = 1e100;
  MaxExt[0] = MaxExt[1] = MaxExt[2] = -1e100;
  if (view_data->frame->selection == NULL)
    return;
  for (tmp_select = view_data->frame->selection;
       tmp_select ;
       tmp_select = g_list_next(tmp_select) )
  {
    object = tmp_select->data;
    /* Find the extents */
    for (TmpTri = object->FirstTriangle ;
         TmpTri ;
         TmpTri = TmpTri->Next)
    {
      if (MinExt[0] > TmpTri->P1[0]) MinExt[0] = TmpTri->P1[0];
      if (MinExt[0] > TmpTri->P2[0]) MinExt[0] = TmpTri->P2[0];
      if (MinExt[0] > TmpTri->P3[0]) MinExt[0] = TmpTri->P3[0];
      if (MinExt[1] > TmpTri->P1[1]) MinExt[1] = TmpTri->P1[1];
      if (MinExt[1] > TmpTri->P2[1]) MinExt[1] = TmpTri->P2[1];
      if (MinExt[1] > TmpTri->P3[1]) MinExt[1] = TmpTri->P3[1];
      if (MinExt[2] > TmpTri->P1[2]) MinExt[2] = TmpTri->P1[2];
      if (MinExt[2] > TmpTri->P2[2]) MinExt[2] = TmpTri->P2[2];
      if (MinExt[2] > TmpTri->P3[2]) MinExt[2] = TmpTri->P3[2];

      if (MaxExt[0] < TmpTri->P1[0]) MaxExt[0] = TmpTri->P1[0];
      if (MaxExt[0] < TmpTri->P2[0]) MaxExt[0] = TmpTri->P2[0];
      if (MaxExt[0] < TmpTri->P3[0]) MaxExt[0] = TmpTri->P3[0];
      if (MaxExt[1] < TmpTri->P1[1]) MaxExt[1] = TmpTri->P1[1];
      if (MaxExt[1] < TmpTri->P2[1]) MaxExt[1] = TmpTri->P2[1];
      if (MaxExt[1] < TmpTri->P3[1]) MaxExt[1] = TmpTri->P3[1];
      if (MaxExt[2] < TmpTri->P1[2]) MaxExt[2] = TmpTri->P1[2];
      if (MaxExt[2] < TmpTri->P2[2]) MaxExt[2] = TmpTri->P2[2];
      if (MaxExt[2] < TmpTri->P3[2]) MaxExt[2] = TmpTri->P3[2];
    }
  }
  switch (view_data->camera_style)
  {
    case ORTHO_XY_CAMERA:
      ZoomX = canvas->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = canvas->allocation.height / (MaxExt[1]-MinExt[1]);
      view_data->x_off = (MaxExt[0]+MinExt[0]) / 2.0;
      view_data->y_off = (MaxExt[1]+MinExt[1]) / 2.0;
      break;

    case ORTHO_XZ_CAMERA:
      ZoomX = canvas->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = canvas->allocation.height / (MaxExt[2]-MinExt[2]);
      view_data->x_off = (MaxExt[0]+MinExt[0]) / 2.0;
      view_data->y_off = (MaxExt[2]+MinExt[2]) / 2.0;
      break;

    case ORTHO_ZY_CAMERA:
      ZoomX = canvas->allocation.width / (MaxExt[2]-MinExt[2]);
      ZoomY = canvas->allocation.height / (MaxExt[1]-MinExt[1]);
      view_data->x_off = (MaxExt[2]+MinExt[2]) / 2.0;
      view_data->y_off = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
  }
  view_data->zoom = MIN(ZoomX, ZoomY) * 0.95;
  gtk_widget_queue_draw(canvas);
}

/*****************************************************************************
*  FitToSceneFunc
******************************************************************************/
void FitToSceneFunc(void)
{
  ViewStruct         *view_data;
  GSList             *tmp_list;
  ObjectStruct       *TmpObject;
  TriangleListStruct *TmpTri;
  Vector              MaxExt, MinExt;
  double              ZoomX=1.0, ZoomY=1.0;
  GtkWidget          *canvas;

  view_data = get_current_view_data();
  if (!view_data)
    return;
  canvas = view_data->drawing_area;

  MinExt[0] = MinExt[1] = MinExt[2] = 1e100;
  MaxExt[0] = MaxExt[1] = MaxExt[2] = -1e100;
  if (!(view_data->frame->all_objects))
    return;
  for (tmp_list = view_data->frame->all_objects;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    TmpObject = tmp_list->data;
    /* Find the extents */
    for (TmpTri = TmpObject->FirstTriangle ;
         TmpTri ;
         TmpTri = TmpTri->Next)
    {
      if (MinExt[0] > TmpTri->P1[0]) MinExt[0] = TmpTri->P1[0];
      if (MinExt[0] > TmpTri->P2[0]) MinExt[0] = TmpTri->P2[0];
      if (MinExt[0] > TmpTri->P3[0]) MinExt[0] = TmpTri->P3[0];
      if (MinExt[1] > TmpTri->P1[1]) MinExt[1] = TmpTri->P1[1];
      if (MinExt[1] > TmpTri->P2[1]) MinExt[1] = TmpTri->P2[1];
      if (MinExt[1] > TmpTri->P3[1]) MinExt[1] = TmpTri->P3[1];
      if (MinExt[2] > TmpTri->P1[2]) MinExt[2] = TmpTri->P1[2];
      if (MinExt[2] > TmpTri->P2[2]) MinExt[2] = TmpTri->P2[2];
      if (MinExt[2] > TmpTri->P3[2]) MinExt[2] = TmpTri->P3[2];

      if (MaxExt[0] < TmpTri->P1[0]) MaxExt[0] = TmpTri->P1[0];
      if (MaxExt[0] < TmpTri->P2[0]) MaxExt[0] = TmpTri->P2[0];
      if (MaxExt[0] < TmpTri->P3[0]) MaxExt[0] = TmpTri->P3[0];
      if (MaxExt[1] < TmpTri->P1[1]) MaxExt[1] = TmpTri->P1[1];
      if (MaxExt[1] < TmpTri->P2[1]) MaxExt[1] = TmpTri->P2[1];
      if (MaxExt[1] < TmpTri->P3[1]) MaxExt[1] = TmpTri->P3[1];
      if (MaxExt[2] < TmpTri->P1[2]) MaxExt[2] = TmpTri->P1[2];
      if (MaxExt[2] < TmpTri->P2[2]) MaxExt[2] = TmpTri->P2[2];
      if (MaxExt[2] < TmpTri->P3[2]) MaxExt[2] = TmpTri->P3[2];
    }
  }
  switch (view_data->camera_style)
  {
    case ORTHO_XY_CAMERA:
      ZoomX = canvas->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = canvas->allocation.height / (MaxExt[1]-MinExt[1]);
      view_data->x_off = (MaxExt[0]+MinExt[0]) / 2.0;
      view_data->y_off = (MaxExt[1]+MinExt[1]) / 2.0;
      break;

    case ORTHO_XZ_CAMERA:
      ZoomX = canvas->allocation.width / (MaxExt[0]-MinExt[0]);
      ZoomY = canvas->allocation.height / (MaxExt[2]-MinExt[2]);
      view_data->x_off = (MaxExt[0]+MinExt[0]) / 2.0;
      view_data->y_off = (MaxExt[2]+MinExt[2]) / 2.0;
      break;

    case ORTHO_ZY_CAMERA:
      ZoomX = canvas->allocation.width / (MaxExt[2]-MinExt[2]);
      ZoomY = canvas->allocation.height / (MaxExt[1]-MinExt[1]);
      view_data->x_off = (MaxExt[2]+MinExt[2]) / 2.0;
      view_data->y_off = (MaxExt[1]+MinExt[1]) / 2.0;
      break;
  }
  view_data->zoom = MIN(ZoomX, ZoomY) * 0.95;
  gtk_widget_queue_draw(canvas);
}

/*****************************************************************************
*  SelectAll
******************************************************************************/
static void SelectAll(void)
{
  ViewStruct   *view_data;
  GSList       *tmp_list;
  ObjectStruct *tmp_object;
  ViewStruct   *TmpView;

  view_data = get_current_view_data();
  DestroySelection(&(view_data->frame->selection));

  for (tmp_list = view_data->frame->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    tmp_object = tmp_list->data;
    view_data->frame->selection = g_list_append(view_data->frame->selection,
                                                tmp_object);
    tmp_object->selected = TRUE;
    CSGTreeSelect(tmp_object);
  }

  /* We redraw all the Views */
  for (tmp_list = view_data->frame->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
}

/*****************************************************************************
*  SelectNone
******************************************************************************/
static void SelectNone(void)
{
  ViewStruct *view_data;
  GSList     *tmp_list;
  ViewStruct *TmpView;

  view_data = get_current_view_data();
  DestroySelection(&(view_data->frame->selection));

  /* We redraw all the Views */
  for (tmp_list = view_data->frame->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
}

/*****************************************************************************
*  InvertSelection
******************************************************************************/
static void InvertSelection(void)
{
  ViewStruct      *view_data;
  ViewStruct      *TmpView;
  GList           *tmp_selection;
  GList           *Prev;
  GSList          *tmp_list;
  ObjectStruct    *tmp_object;

  view_data = get_current_view_data();
  tmp_selection = NULL;
  for (tmp_list = view_data->frame->all_objects ;
       tmp_list ;
       tmp_list = g_slist_next(tmp_list))
  {
    tmp_object = tmp_list->data;

    tmp_selection = g_list_append(tmp_selection, tmp_object);
    tmp_object->selected = TRUE;
    CSGTreeSelect(tmp_object);
  }

  if (view_data->frame->selection)
  {
    for (Prev = view_data->frame->selection ;
         Prev ;
         Prev = g_list_next(Prev) )
    {
      tmp_object = Prev->data;
      tmp_selection = g_list_remove(tmp_selection, tmp_selection);
      tmp_object->selected = FALSE;
      CSGTreeUnselect(tmp_object);
    }
  }
  DestroySelection(&(view_data->frame->selection));
  view_data->frame->selection = tmp_selection;
  /* Update the CSG Tree selection. A little drastic, perhaps, but it'll
   * do until someone comes up with a better way. */
  RebuildCSGTree(NULL, view_data->frame);
  /* We redraw all the Views */
  for (tmp_list = view_data->frame->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
}

/*****************************************************************************
*  TranslateSelection
******************************************************************************/
static void TranslateSelection(GtkWidget *Dialog)
{
  Vector            Vect;
  GtkWidget       **Data;
  ViewStruct       *view_data;
  GList            *selection;
  ObjectStruct     *object;
  GSList           *tmp_list;
  ViewStruct       *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  view_data = get_current_view_data();

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (view_data->frame->selection)
  {
    for (selection = view_data->frame->selection ;
         selection ;
         selection = g_list_next(selection) )
    {
      object = selection->data;
      giram_object_translate(object, Vect);
      giram_object_build_triangle_mesh(object);
    }
    for (tmp_list = view_data->frame->all_views ;
         tmp_list ;
         tmp_list = tmp_list->next)
    {
      TmpView = tmp_list->data;
      gtk_widget_queue_draw(TmpView->drawing_area);
    }
  }
  gtk_widget_destroy(Dialog);
}

/*****************************************************************************
*  RotateSelection
******************************************************************************/
static void RotateSelection(GtkWidget *Dialog)
{
  Vector            Vect;
  GtkWidget       **Data;
  ViewStruct       *view_data;
  GList            *selection;
  ObjectStruct     *object;
  GSList           *tmp_list;
  ViewStruct       *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  view_data = get_current_view_data();

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (view_data->frame->selection)
  {
    for (selection = view_data->frame->selection ;
         selection ;
         selection = g_list_next(selection) )
    {
      object = selection->data;
      giram_object_rotate(object, Vect);
      giram_object_build_triangle_mesh(object);
    }
    for (tmp_list = view_data->frame->all_views ;
         tmp_list ;
         tmp_list = tmp_list->next)
    {
      TmpView = tmp_list->data;
      gtk_widget_queue_draw(TmpView->drawing_area);
    }
  }
  gtk_widget_destroy(Dialog);
}

/*****************************************************************************
*  ScaleSelection
******************************************************************************/
static void ScaleSelection(GtkWidget *Dialog)
{
  Vector            Vect;
  GtkWidget       **Data;
  ViewStruct       *view_data;
  GList            *selection;
  ObjectStruct     *object;
  GSList           *tmp_list;
  ViewStruct       *TmpView;

  Data = gtk_object_get_data(GTK_OBJECT(Dialog), "WidgetData");
  view_data = get_current_view_data();

  Vect[0] = atof(gtk_entry_get_text(GTK_ENTRY(Data[0])));
  Vect[1] = atof(gtk_entry_get_text(GTK_ENTRY(Data[1])));
  Vect[2] = atof(gtk_entry_get_text(GTK_ENTRY(Data[2])));
  if (fabs(Vect[0]) < 10e-8)
  {
    Vect[0] = 1.0;
    g_message(_("You can't scale by 0, changed to 1"));
  }
  if (fabs(Vect[1]) < 10e-8)
  {
    Vect[1] = 1.0;
    g_message(_("You can't scale by 0, changed to 1"));
  }
  if (fabs(Vect[2]) < 10e-8)
  {
    Vect[2] = 1.0;
    g_message(_("You can't scale by 0, changed to 1"));
  }
  if (view_data->frame->selection)
  {
    for (selection = view_data->frame->selection ;
         selection ;
         selection = g_list_next(selection) )
    {
      object = selection->data;
      giram_object_scale(object, Vect);
      giram_object_build_triangle_mesh(object);
    }
    for (tmp_list = view_data->frame->all_views ;
         tmp_list ;
         tmp_list = tmp_list->next)
    {
      TmpView = tmp_list->data;
      gtk_widget_queue_draw(TmpView->drawing_area);
    }
  }
  gtk_widget_destroy(Dialog);
}

/*****************************************************************************
*  TRSSelectionDialog
******************************************************************************/
static void TRSSelectionDialog(gchar *title, GtkSignalFunc CallBack)
{
  GtkWidget  *dialog, *Table, *Label, *Entry;
  GtkWidget **Data;

  dialog = giram_dialog_new(title, "giram_dialog",
                            NULL /* FIXME */,
                            "dialogs/transforms.html",
                            GTK_WIN_POS_MOUSE,
                            FALSE, FALSE, FALSE,

                            _("OK"), CallBack,
                            NULL, 1, NULL, TRUE, FALSE,

                            _("Cancel"), gtk_widget_destroy,
                            NULL, 1, NULL, FALSE, TRUE,
                            NULL);
  Data = g_new(GtkWidget*, 3);
  gtk_object_set_data(GTK_OBJECT(dialog), "WidgetData", Data);

  Table = gtk_table_new(2, 3, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), Table);
  /* X */
  Label = gtk_label_new(" x: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 0,1,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);

  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 0,1,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  Data[0] = Entry;

  /* Y */
  Label = gtk_label_new(" y: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 1,2,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);

  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 1,2,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  Data[1] = Entry;

  /* Z */
  Label = gtk_label_new(" z: ");
  gtk_table_attach(GTK_TABLE(Table), Label, 0,1, 2,3,
                   GTK_FILL | GTK_EXPAND,
                   GTK_FILL | GTK_EXPAND, 0, 0);

  Entry = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(Table), Entry, 1,2, 2,3,
                   GTK_FILL | GTK_EXPAND,
                   GTK_EXPAND, 0, 0);
  Data[2] = Entry;

  gtk_widget_show_all(dialog);
}

/*****************************************************************************
*  TranslateSelectionDialog
******************************************************************************/
static void TranslateSelectionDialog(void)
{
  TRSSelectionDialog(_("Translate Selection"), TranslateSelection);
}

/*****************************************************************************
*  RotateSelectionDialog
******************************************************************************/
static void RotateSelectionDialog(void)
{
  TRSSelectionDialog(_("Rotate Selection"), RotateSelection);
}

/*****************************************************************************
*  ScaleSelectionDialog
******************************************************************************/
static void ScaleSelectionDialog(void)
{
  TRSSelectionDialog(_("Scale Selection"), ScaleSelection);
}

/*****************************************************************************
*  DrawSelection
******************************************************************************/
void DrawSelection(GtkWidget *DrawingArea, TriangleListStruct *Tri,
                   GdkGC *gc, double Xoff, double Yoff, double Zoom)
{ /* FIXME: OO me */
  ViewStruct *view_data;

  view_data = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  switch (view_data->camera_style)
  {
    case ORTHO_XY_CAMERA:
      DrawSelectionXY(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;

    case ORTHO_XZ_CAMERA:
      DrawSelectionXZ(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;

    case ORTHO_ZY_CAMERA:
      DrawSelectionZY(DrawingArea, Tri, gc, Xoff, Yoff, Zoom);
      break;
  }
}

/*****************************************************************************
*  draw_unselected
******************************************************************************/
void draw_unselected(GdkDrawable *drawable, gint style, gint width, gint height)
{ /* FIXME: OO me */
  switch (style)
  {
    case ORTHO_XY_CAMERA:
      draw_unselected_XY(drawable, width, height);
      break;

    case ORTHO_XZ_CAMERA:
      draw_unselected_XZ(drawable, width, height);
      break;

    case ORTHO_ZY_CAMERA:
      draw_unselected_ZY(drawable, width, height);
      break;
  }
}

/*****************************************************************************
*  draw_all
******************************************************************************/
void draw_all(GdkDrawable *drawable, gint style, gint width, gint height)
{ /* FIXME: OO me */
  switch (style)
  {
    case ORTHO_XY_CAMERA:
      draw_all_XY(drawable, width, height);
      break;

    case ORTHO_XZ_CAMERA:
      draw_all_XZ(drawable, width, height);
      break;

    case ORTHO_ZY_CAMERA:
      draw_all_ZY(drawable, width, height);
      break;
  }
}

/*****************************************************************************
*  event_dispatcher
******************************************************************************/
gint event_dispatcher(GtkWidget *canvas, GdkEvent *event, gpointer Menu)
{
  current_view_data = gtk_object_get_data(GTK_OBJECT(canvas), "ViewData");
  switch (event->type)
  {
    case GDK_BUTTON_PRESS:
    {
      GdkEventButton *bevent = (GdkEventButton *)event;
      if (bevent->button == 3) /* Right Button ==> Popup Menu */
      {
        gtk_menu_popup(GTK_MENU(Menu), NULL, NULL, NULL, NULL,
                                              bevent->button, bevent->time);
      }
      else if (bevent->button == 1)
      { /* Left Button : this is the big game */
        switch (ModeTool)
        {
          case MT_SELECT:       /* The Select Tool */
            ToggleSelect(canvas, bevent);
            break;
          case MT_MOVE:         /* The Move Tool */
            ToolInitMoveSelection(canvas, bevent);
            break;
          case MT_ROTATE:       /* The Rotate Tool */
            ToolInitRotateSelection(canvas, bevent);
            break;
          case MT_SCALE:        /* The Scale Tool */
            ToolInitScaleSelection(canvas, bevent);
            break;
          case MT_NEW_BOX:      /* The New Box Tool */
            ToolInitNewBox(canvas, bevent);
            break;
          case MT_NEW_CONE:     /* The New Cone Tool */
            ToolInitNewCone(canvas, bevent);
            break;
/* FIXME          case MT_NEW_CSG:
            ToolNewCSG(DrawingArea, bevent);
            break;*/
          case MT_NEW_CYLINDER: /* The New Cylinder Tool */
            ToolInitNewCylinder(canvas, bevent);
            break;
          case MT_NEW_DISC:
            ToolInitNewDisc(canvas, bevent);
            break;
          case MT_NEW_PLANE:    /* The New Plane Tool */
            ToolNewPlane(canvas, bevent);
            break;
          case MT_NEW_SOR:
            ToolNewSor(canvas, bevent);
            break;
          case MT_NEW_SPHERE:   /* The New Sphere Tool */
            ToolInitNewSphere(canvas, bevent);
            break;
          case MT_NEW_SUPERELLIPSOID: /* The New SuperEllipsoid Tool */
            ToolNewSuperEllipsoid(canvas, bevent);
            break;
          case MT_NEW_TORUS:    /* The New Torus Tool */
            ToolInitNewTorus(canvas, bevent);
            break;
          case MT_NEW_TRIANGLE:
            ToolNewTriangle(canvas, bevent);
            break;
          case MT_NEW_LIGHTSOURCE:
            ToolNewLightSource(canvas, bevent);
            break;
          case MT_LAST:
            break;
        }
      }
      return TRUE;
    }
    case GDK_MOTION_NOTIFY:
    {
      GtkWidget *status_bar;
      guint      id;
      gchar     *coord_txt = NULL;
      gdouble    xs, ys, xoff, yoff, zoom;

      xoff = current_view_data->x_off;
      yoff = current_view_data->y_off;
      zoom = current_view_data->zoom;
      xs = xoff+(((GdkEventMotion *)event)->x-canvas->allocation.width /2)/zoom;
      ys = yoff-(((GdkEventMotion *)event)->y-canvas->allocation.height/2)/zoom;

      status_bar = current_view_data->status_bar;
      id = gtk_statusbar_get_context_id(GTK_STATUSBAR(status_bar), "mouse coord");
      gtk_statusbar_pop(GTK_STATUSBAR(status_bar), id);
      switch (current_view_data->camera_style)
      {
        case ORTHO_XY_CAMERA:
          coord_txt = g_strdup_printf("<%g, %g, 0>", xs, ys);
          break;
        case ORTHO_XZ_CAMERA:
          coord_txt = g_strdup_printf("<%g, 0, %g>", xs, ys);
          break;
        case ORTHO_ZY_CAMERA:
          coord_txt = g_strdup_printf("<0, %g, %g>", ys, xs);
          break;
      }
      gtk_statusbar_push(GTK_STATUSBAR(status_bar), id, coord_txt);  
      g_free(coord_txt);

      switch (ModeTool)
      {
        case MT_MOVE:
          ToolMotionMoveSelection(canvas, event);
          break;
        case MT_ROTATE:
          ToolMotionRotateSelection(canvas, event);
          break;
        case MT_SCALE:
          ToolMotionScaleSelection(canvas, event);
          break;
        case MT_NEW_BOX:
          ToolMotionNewBox(canvas, event);
          break;
        case MT_NEW_CONE:
          ToolMotionNewCone(canvas, event);
          break;
        case MT_NEW_CYLINDER:
          ToolMotionNewCylinder(canvas, event);
          break;
        case MT_NEW_DISC:
          ToolMotionNewDisc(canvas, event);
          break;
        case MT_NEW_SPHERE:
          ToolMotionNewSphere(canvas, event);
          break;
        case MT_NEW_TORUS:
          ToolMotionNewTorus(canvas, event);
          break;
        default:
          break;
      }
      return FALSE;
    }
    case GDK_BUTTON_RELEASE:
    {
      switch (ModeTool)
      {
        case MT_MOVE:
          ToolReleaseMoveSelection(canvas, event);
          break;
        case MT_ROTATE:
          ToolReleaseRotateSelection(canvas, event);
          break;
        case MT_SCALE:
          ToolReleaseScaleSelection(canvas, event);
          break;
        case MT_NEW_BOX:
          ToolReleaseNewBox(canvas, event);
          break;
        case MT_NEW_CONE:
          ToolReleaseNewCone(canvas, event);
          break;
        case MT_NEW_CYLINDER:
          ToolReleaseNewCylinder(canvas, event);
          break;
        case MT_NEW_DISC:
          ToolReleaseNewDisc(canvas, event);
          break;
        case MT_NEW_SPHERE:
          ToolReleaseNewSphere(canvas, event);
          break;
        case MT_NEW_TORUS:
          ToolReleaseNewTorus(canvas, event);
          break;
        default:
          break;
      }
      return TRUE;
    }
    case GDK_KEY_PRESS:
      g_print("a key press event has been catched...\n");
      break;
    case GDK_KEY_RELEASE:
      g_print("a key release event has been catched...\n");
      break;
    default:
      return FALSE;
  }
  return FALSE;
}

/*****************************************************************************
*  HRulerConfigure
******************************************************************************/
void HRulerConfigure(GtkWidget *DrawingArea, gpointer data, GtkWidget *HRuler)
{
  ViewStruct *view_data;

  view_data = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_ruler_set_range(GTK_RULER(HRuler),
                      view_data->x_off-DrawingArea->allocation.width / 2.0 / view_data->zoom,
                      view_data->x_off+DrawingArea->allocation.width / 2.0 / view_data->zoom,
                      0.0, 50.0);
}

/*****************************************************************************
*  VRulerConfigure
******************************************************************************/
void VRulerConfigure(GtkWidget *DrawingArea, gpointer data, GtkWidget *VRuler)
{
  ViewStruct *view_data;

  view_data = gtk_object_get_data(GTK_OBJECT(DrawingArea), "ViewData");
  gtk_ruler_set_range(GTK_RULER(VRuler),
                      view_data->y_off-DrawingArea->allocation.height / 2.0 / view_data->zoom,
                      view_data->y_off+DrawingArea->allocation.height / 2.0 / view_data->zoom,
                      0.0, 50.0);
}

/*****************************************************************************
*  gview_delete
******************************************************************************/
static gint gview_delete(GtkWidget *window, GdkEvent *ev, ViewStruct *FrameView)
{
  g_print("The deletion of an XYZ view has been catched!\n");
  FrameView->frame->all_views = g_slist_remove(FrameView->frame->all_views, FrameView);
  g_free(FrameView);
  return FALSE;
}

static GtkItemFactoryEntry ViewMenu[] =
{
  {"/_File",                         NULL, NULL,                     0, "<Branch>"},
  {"/File/Load _Pov",                NULL, LoadPovFunc,              0, "<Item>"},
  {"/File/Load AutoCAD _DXF",        NULL, LoadDXFFunc,              0, "<Item>"},
#ifdef USE_LIB3DS
  {"/File/Load 3DStudio _3DS",       NULL, Load3DSFunc,              0, "<Item>"},
#endif
  {"/File/Load GTS",                 NULL, LoadGTSFunc,              0, "<Item>"},
  {"/File/Dum1",                     NULL, NULL,                     0, "<Separator>"},
  {"/File/Save P_ov",                NULL, SavePovFunc,              0, "<Item>"},
  {"/File/Save AutoCAD D_XF",        NULL, SaveDXFFunc,              0, "<Item>"},
#ifdef WITH_S3D
  {"/File/Save _Bishop S3D",         NULL, SaveS3DFunc,              0, "<Item>"},
#endif
  {"/File/Dum2",                     NULL, NULL,                     0, "<Separator>"},
  {"/File/_Quit",                    NULL, ExitFunc,                 0, "<Item>"},
  {"/_Edit",                         NULL, NULL,                     0, "<Branch>"},
  {"/Edit/_Cut",                     NULL, CutSelection,             0, "<Item>"},
  {"/Edit/C_opy",                    NULL, CopySelection,            0, "<Item>"},
  {"/Edit/_Paste",                   NULL, PasteSelection,           0, "<Item>"},
  {"/Edit/_Snap",                    NULL, SnapDialogShow,           0, "<Item>"},
  {"/_View",                         NULL, NULL,                     0, "<Branch>"},
  {"/View/Zoom _in",                 NULL, ZoomIn,                   0, "<Item>"},
  {"/View/Zoom _out",                NULL, ZoomOut,                  0, "<Item>"},
  {"/View/Fit To Se_lection",        NULL, FitToSelectionFunc,       0, "<Item>"},
  {"/View/Fit To _Scene",            NULL, FitToSceneFunc,           0, "<Item>"},
  {"/_Selection",                    NULL, NULL,                     0, "<Branch>"},
  {"/Selection/Select _All",         NULL, SelectAll,                0, "<Item>"},
  {"/Selection/Select _None",        NULL, SelectNone,               0, "<Item>"},
  {"/Selection/_Invert Selection",   NULL, InvertSelection,          0, "<Item>"},
  {"/Selection/Export to _DXF",      NULL, SaveSelectedDXFFunc,      0, "<Item>"},
#ifdef WITH_S3D
  {"/Selection/Export to _S3D",      NULL, SaveSelectedS3DFunc,      0, "<Item>"},
#endif
  {"/Selection/Import DXF _Objects", NULL, InsertDXFFunc,            0, "<Item>"},
#ifdef USE_LIB3DS
  {"/Selection/Import 3DS Objects",  NULL, Insert3DSFunc,            0, "<Item>"},
#endif
  {"/Selection/Import GTS Objects",  NULL, InsertGTSFunc,            0, "<Item>"},
  {"/_Operations",                   NULL, NULL,                     0, "<Branch>"},
  {"/Operations/_Translate",         NULL, TranslateSelectionDialog, 0, "<Item>"},
  {"/Operations/_Rotate",            NULL, RotateSelectionDialog,    0, "<Item>"},
  {"/Operations/_Scale",             NULL, ScaleSelectionDialog,     0, "<Item>"},
  {"/Operations/Change _Pigment",    NULL, ChangeColorDialog,        0, "<Item>"},
  {"/Operations/_Edit Camera",       NULL, EditCamera,               0, "<Item>"},
  {"/_Render",                       NULL, NULL,                     0, "<Branch>"},
  {"/Render/_Wire Frame",            NULL, SetWireFrameExpose,       0, "<Item>"},
  {"/Render/_Hidden Faces",          NULL, SetHiddenFacesExpose,     0, "<Item>"},
  {"/Render/_Gouraud",               NULL, SetGouraudExpose,         0, "<Item>"},
  {"/Render/_PovRay Rendering",      NULL, CallPovFront,             0, "<Item>"},
  {"/_Dialogs",                      NULL, NULL,                     0, "<Branch>"},
  {"/Dialogs/_CSG Tree",             NULL, ShowCSGTree,              0, "<Item>"},
  {"/Dialogs/_Tool Options",         NULL, ShowToolOptionsDialog,    0, "<Item>"},
  {"/_Plug-Ins",                     NULL, NULL,                     0, "<Branch>"}
};
static guint NbViewEntries = sizeof(ViewMenu) /
                               sizeof(GtkItemFactoryEntry);

static GtkAccelGroup *ViewAccelGroup;
static GtkItemFactory *ViewFactory;


/*****************************************************************************
*  CreateViewMenu
******************************************************************************/
void CreateViewMenu(void)
{
  /* New Keyboard accel group */
  ViewAccelGroup = gtk_accel_group_new();

  /* Creation of the item factory */
  ViewFactory = gtk_item_factory_new(GTK_TYPE_MENU, "<View>", ViewAccelGroup);
  gtk_item_factory_create_items(ViewFactory, NbViewEntries, ViewMenu, 0);
}

gboolean gdisplay_origin_button_press(GtkWidget      *widget,
                                      GdkEventButton *event,
                                      gpointer        data)
{
#if 0
  GDisplay *gdisp;

  gdisp = (GDisplay *) data;

  if (! gimp_busy && event->button == 1)
  {
    gint x, y;

    gdisplay_origin_menu_position(GTK_MENU(gdisp->ifactory->widget),
                                  &x, &y, widget);
    gtk_item_factory_popup_with_data(gdisp->ifactory,
                                     gdisp->gimage, NULL,
                                     x, y,
                                     1, event->time);
  }

  /*  Stop the signal emission so the button doesn't grab the
   *  pointer away from us
   */
  gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_press_event");
#endif
  return FALSE;
}



#define CANVAS_EVENT_MASK (GDK_EXPOSURE_MASK            | \
                           GDK_POINTER_MOTION_MASK      | \
                           GDK_POINTER_MOTION_HINT_MASK | \
                           GDK_BUTTON_PRESS_MASK        | \
                           GDK_BUTTON_RELEASE_MASK      | \
                           GDK_STRUCTURE_MASK           | \
                           GDK_ENTER_NOTIFY_MASK        | \
                           GDK_LEAVE_NOTIFY_MASK        | \
                           GDK_KEY_PRESS_MASK           | \
                           GDK_KEY_RELEASE_MASK         | \
                           GDK_PROXIMITY_OUT_MASK)

/*****************************************************************************
*  NewView
******************************************************************************/
void NewView(FrameStruct *LocalFrame, int CameraStyle)
{
  static GdkPixmap *navbutton_pixmap  = NULL;
  static GdkBitmap *navbutton_mask    = NULL;

  GtkWidget *window, *main_vbox, *disp_vbox, *upper_hbox, *inner_table;
  GtkWidget *lower_hbox, *hsb, *vsb;
  GtkWidget *origin, *arrow, *hrule, *vrule;
  GtkAdjustment *hsbdata, *vsbdata;
  GtkWidget *canvas, *nav_ebox;

  GtkWidget *statusarea, *status_hbox, *statusbar;
  GtkWidget *progressbar, *cancelbutton;

  GtkWidget *pixmap;

  GtkWidget *ruler;

  gint contextid;

  /*  GtkWidget *Arrow; */
  GtkWidget  *DrawingArea;
  ViewStruct *view_data;
  GtkWidget  *menu;

  view_data = g_new(ViewStruct, 1);

  LocalFrame->all_views = g_slist_append(LocalFrame->all_views, view_data);

  view_data->frame = LocalFrame;
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);

  gtk_window_set_wmclass(GTK_WINDOW(window), "view_window", "Giram");
  gtk_widget_set_usize(window, 512, 384); /* XXX: should be setable */
  gtk_widget_set_events(window, GDK_POINTER_MOTION_MASK
                              | GDK_POINTER_MOTION_HINT_MASK
                              | GDK_BUTTON_PRESS_MASK
                              | GDK_KEY_PRESS_MASK
                              | GDK_KEY_RELEASE_MASK);
  gtk_container_set_border_width(GTK_CONTAINER(window), 0);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
                     (GtkSignalFunc)gview_delete, view_data);

  /*  main_vbox
   *     |
   *     +-- disp_vbox
   *     |      |
   *     |      +-- upper_hbox
   *     |      |      |
   *     |      |      +-- inner_table
   *     |      |      |      |
   *     |      |      |      +-- origin
   *     |      |      |      +-- hruler
   *     |      |      |      +-- vruler
   *     |      |      |      +-- canvas
   *     |      |      |
   *     |      |      +-- vscrollbar
   *     |      |
   *     |      +-- lower_hbox
   *     |             |
   *     |             +-- qmaskoff
   *     |             +-- qmaskon
   *     |             +-- hscrollbar
   *     |             +-- navbutton
   *     |
   *     +-- statusarea
   *            |
   *            +-- cursorlabel
   *            +-- statusbar
   *            +-- progressbar
   *            +-- cancelbutton
   */

  /*  first, set up the container hierarchy  *********************************/

  /*  the vbox containing all widgets  */
  main_vbox = gtk_vbox_new(FALSE, 2);
  gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 2);
  gtk_container_add(GTK_CONTAINER(window), main_vbox);

  /*  another vbox for everything except the statusbar  */
  disp_vbox = gtk_vbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX(main_vbox), disp_vbox, TRUE, TRUE, 0);
  gtk_widget_show(disp_vbox);

  /*  a hbox for the inner_table and the vertical scrollbar  */
  upper_hbox = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX(disp_vbox), upper_hbox, TRUE, TRUE, 0);
  gtk_widget_show(upper_hbox);

  /*  the table containing origin, rulers and the canvas  */
  inner_table = gtk_table_new(2, 2, FALSE);
  gtk_table_set_col_spacing(GTK_TABLE(inner_table), 0, 1);
  gtk_table_set_row_spacing(GTK_TABLE(inner_table), 0, 1);
  gtk_box_pack_start(GTK_BOX(upper_hbox), inner_table, TRUE, TRUE, 0);
  gtk_widget_show(inner_table);

  /*  the hbox containing qmask buttons, vertical scrollbar and nav button  */
  lower_hbox = gtk_hbox_new(FALSE, 1);
  gtk_box_pack_start(GTK_BOX(disp_vbox), lower_hbox, FALSE, FALSE, 0);
  gtk_widget_show(lower_hbox);

  /*  eventbox and hbox for status area  */
  /*gdisp->*/statusarea = gtk_event_box_new();
  gtk_box_pack_start(GTK_BOX(main_vbox), /*gdisp->*/statusarea, FALSE, FALSE, 0);

  status_hbox = gtk_hbox_new(FALSE, 2);
  gtk_container_add(GTK_CONTAINER(/*gdisp->*/statusarea), status_hbox);
  gtk_widget_show(status_hbox);

  gtk_container_set_resize_mode(GTK_CONTAINER(status_hbox),
                                GTK_RESIZE_QUEUE);

  /*  create the scrollbars  *************************************************/

  /*  the horizontal scrollbar  */
  /*gdisp->*/hsbdata = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 512/*width*/, 1, 1, 512/*width*/));
  /*gdisp->*/hsb = gtk_hscrollbar_new(/*gdisp->*/hsbdata);
  GTK_WIDGET_UNSET_FLAGS(/*gdisp->*/hsb, GTK_CAN_FOCUS);

  /*  the vertical scrollbar  */
  /*gdisp->*/vsbdata = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 384/*height*/, 1, 1, 384/*height*/));
  /*gdisp->*/vsb = gtk_vscrollbar_new(/*gdisp->*/vsbdata);
  GTK_WIDGET_UNSET_FLAGS(/*gdisp->*/vsb, GTK_CAN_FOCUS);

  /*  create the contents of the inner_table  ********************************/

  /*  the menu popup button  */
  /*gdisp->*/origin = gtk_button_new();
  GTK_WIDGET_UNSET_FLAGS(/*gdisp->*/origin, GTK_CAN_FOCUS);
  gtk_widget_set_events(GTK_WIDGET(/*gdisp->*/origin),
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
#if 0
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/origin), "button_press_event",
                     (GtkSignalFunc)gview_origin_button_press, /*gdisp*/NULL);
#endif

  arrow = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_OUT);
  gtk_container_set_border_width(GTK_CONTAINER(/*gdisp->*/origin), 0);
  gtk_container_add(GTK_CONTAINER(/*gdisp->*/origin), arrow);
  gtk_widget_show(arrow);

  /*  the horizontal ruler  */
  /*gdisp->*/hrule = gtk_hruler_new();
  gtk_widget_set_events(/*gdisp->*/hrule,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  gtk_signal_connect_object(GTK_OBJECT(window), "motion_notify_event",
                            (GtkSignalFunc)GTK_WIDGET_CLASS(GTK_OBJECT(/*gdisp->*/hrule)->klass)->motion_notify_event,
                            GTK_OBJECT(/*gdisp->*/hrule));
#if 0
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/hrule), "button_press_event",
                     (GtkSignalFunc)gview_hruler_button_press, /*gdisp*/NULL);
#endif

  /*  the vertical ruler  */
  /*gdisp->*/vrule = gtk_vruler_new();
  gtk_widget_set_events(/*gdisp->*/vrule,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  gtk_signal_connect_object(GTK_OBJECT(window), "motion_notify_event",
                            (GtkSignalFunc)GTK_WIDGET_CLASS(GTK_OBJECT(/*gdisp->*/vrule)->klass)->motion_notify_event,
                            GTK_OBJECT(/*gdisp->*/vrule));
#if 0
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/vrule), "button_press_event",
                     (GtkSignalFunc)gdisplay_vruler_button_press, /*gdisp*/NULL);
#endif

  /*  the canvas  */
  /*gdisp->*/canvas = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(/*gdisp->*/canvas), 512, 384);
  gtk_widget_set_events(/*gdisp->*/canvas, CANVAS_EVENT_MASK);
  gtk_widget_set_extension_events(/*gdisp->*/canvas, GDK_EXTENSION_EVENTS_ALL);
  GTK_WIDGET_SET_FLAGS(/*gdisp->*/canvas, GTK_CAN_FOCUS);
/*  gtk_object_set_user_data(GTK_OBJECT(gdisp->canvas), (gpointer)gdisp);*/

  /*  set the active display before doing any other canvas event processing  */
#if 0
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/canvas), "event",
                     (GtkSignalFunc)gdisplay_shell_events, gdisp);
  gtk_signal_connect(GTK_OBJECT(/*gdisp->*/canvas), "event",
                     GTK_SIGNAL_FUNC (gdisplay_canvas_events), gdisp);
#endif

  /*  the navigation window button  */
  nav_ebox = gtk_event_box_new();
#if 0
  gtk_signal_connect(GTK_OBJECT(nav_ebox), "button_press_event",
                     (GtkSignalFunc)nav_popup_click_handler, /*gdisp*/NULL);
#endif

  /* We need to realize the shell so that we have a GdkWindow for
   * the pixmap creation.  */
  gtk_widget_realize(window);

  /*  create the pixmaps  ****************************************************/
  if (!navbutton_pixmap)
  {
    GtkStyle *style;

    style = gtk_widget_get_style(window);

    navbutton_pixmap = gdk_pixmap_create_from_xpm_d(window->window,
                                                    &navbutton_mask,
                                                    &style->bg[GTK_STATE_NORMAL],
                                                    navbutton_xpm);
  }

  /*  create the GtkPixmaps  */
  pixmap = gtk_pixmap_new(navbutton_pixmap, navbutton_mask);
  gtk_container_add(GTK_CONTAINER(nav_ebox), pixmap);
  gtk_widget_show(pixmap);

  /*  create the contents of the status area *********************************/

  /*  the statusbar  */
  /*gdisp->*/statusbar = gtk_statusbar_new();
  gtk_widget_set_usize(/*gdisp->*/statusbar, 1, -1);
  gtk_container_set_resize_mode(GTK_CONTAINER(/*gdisp->*/statusbar),
                                GTK_RESIZE_QUEUE);
  contextid = gtk_statusbar_get_context_id(GTK_STATUSBAR(/*gdisp->*/statusbar),
                                           "title");
  gtk_statusbar_push(GTK_STATUSBAR(/*gdisp->*/statusbar),
                     contextid,
                     "");

  /*  the progress bar  */
  /*gdisp->*/progressbar = gtk_progress_bar_new();
  gtk_widget_set_usize(/*gdisp->*/progressbar, 80, -1);

  /*  the cancel button  */
  /*gdisp->*/cancelbutton = gtk_button_new_with_label(_("Cancel"));
  gtk_widget_set_sensitive(/*gdisp->*/cancelbutton, FALSE);

  /*  pack all the widgets  **************************************************/

  /*  fill the upper_hbox  */
  gtk_box_pack_start(GTK_BOX(upper_hbox), /*gdisp->*/vsb, FALSE, FALSE, 0);

  /*  fill the inner_table  */
  gtk_table_attach(GTK_TABLE(inner_table), /*gdisp->*/origin, 0, 1, 0, 1,
                   GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(inner_table), /*gdisp->*/hrule, 1, 2, 0, 1,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(inner_table), /*gdisp->*/vrule, 0, 1, 1, 2,
                   GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(inner_table), /*gdisp->*/canvas, 1, 2, 1, 2,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);

  /*  fill the lower_hbox  */
  gtk_box_pack_start(GTK_BOX(lower_hbox), /*gdisp->*/hsb, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(lower_hbox), nav_ebox, FALSE, FALSE, 0);

  /*  fill the status area  */
  gtk_box_pack_start(GTK_BOX(status_hbox), /*gdisp->*/statusbar, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(status_hbox), /*gdisp->*/progressbar, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(status_hbox), /*gdisp->*/cancelbutton, FALSE, FALSE, 0);

  /*  show everything  *******************************************************/

/*  if (giramrc.show_rulers)
  {*/
    gtk_widget_show(/*gdisp->*/origin);
    gtk_widget_show(/*gdisp->*/hrule);
    gtk_widget_show(/*gdisp->*/vrule);
/*  }*/
  gtk_widget_show(/*gdisp->*/canvas);

  gtk_widget_show(/*gdisp->*/vsb);
  gtk_widget_show(/*gdisp->*/hsb);

  gtk_widget_show(nav_ebox);

  gtk_widget_show(/*gdisp->*/statusbar);
  gtk_widget_show(/*gdisp->*/progressbar);
  gtk_widget_show(/*gdisp->*/cancelbutton);
/*  if (giramrc.show_statusbar)
  {*/
    gtk_widget_show(/*gdisp->*/statusarea);
/*  }*/

  gtk_widget_realize(/*gdisp->*/canvas);
  gdk_window_set_back_pixmap(/*gdisp->*/canvas->window, NULL, FALSE);

  gtk_widget_show(main_vbox);
  gtk_widget_show(window);

  /*  set the focus to the canvas area  */
  gtk_widget_grab_focus(/*gdisp->*/canvas);

  DrawingArea = gtk_drawing_area_new();

  view_data->drawing_area = canvas;
  view_data->camera_style = CameraStyle;

  switch (CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXY, NULL);
      break;

    case ORTHO_XZ_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewXZ, NULL);
      break;

    case ORTHO_ZY_CAMERA:
      view_data->render_style =
        gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                           (GtkSignalFunc)ExposeWireFrameViewZY, NULL);
      break;
  }
  view_data->zoom = 50.0;
  view_data->x_off = 0.0;
  view_data->y_off = 0.0;
  gtk_object_set_data(GTK_OBJECT(window), "ViewData", view_data);
  gtk_object_set_data(GTK_OBJECT(canvas), "ViewData", view_data);
  current_view_data = view_data;
  view_data->status_bar = statusbar;

  giram_view_set_title(view_data);

  gtk_accel_group_attach(ViewAccelGroup, GTK_OBJECT(window));

  /* We get the menu from the factory */
  menu = gtk_item_factory_get_widget(ViewFactory, "<View>");

  /* The "Plugins" SubMenu */
/*  gtk_menu_append(GTK_MENU(menu), CreatePlugInsMenu(DrawingArea));*/
  gtk_signal_connect(GTK_OBJECT(canvas), "event",
                     GTK_SIGNAL_FUNC(event_dispatcher), GTK_OBJECT(menu));
}

